package ru.dbotthepony.mc.otm

import net.minecraftforge.common.ForgeConfigSpec
import net.minecraftforge.fml.ModLoadingContext
import net.minecraftforge.fml.config.ModConfig
import ru.dbotthepony.mc.otm.block.entity.AndroidStationBlockEntity
import ru.dbotthepony.mc.otm.block.entity.ChemicalGeneratorBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterBottlerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterDecomposerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterRecyclerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterReplicatorBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterScannerBlockEntity
import ru.dbotthepony.mc.otm.capability.BlockEnergyStorageImpl
import ru.dbotthepony.mc.otm.core.Decimal
import ru.dbotthepony.mc.otm.core.defineDecimal
import ru.dbotthepony.mc.otm.item.EnergySwordItem
import ru.dbotthepony.mc.otm.registry.MNames

interface VerboseBalanceValues {
	val capacity: Decimal
	val receive: Decimal
	val extract: Decimal
}

interface BatteryBalanceValues : VerboseBalanceValues {
	val initialBatteryLevel: Decimal
}

interface ConciseBalanceValues {
	val capacity: Decimal
	val throughput: Decimal
}

object ServerConfig {
	private val specBuilder = ForgeConfigSpec.Builder()
	private val spec: ForgeConfigSpec
	private var registered = false

	private fun verboseValues(name: String, storage: Decimal, receive: Decimal, extract: Decimal = receive): VerboseBalanceValues {
		specBuilder.push(name)

		val obj = object : VerboseBalanceValues {
			override val capacity: Decimal by specBuilder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val receive: Decimal by specBuilder.defineDecimal("receive", receive, minimum = Decimal.ONE)
			override val extract: Decimal by specBuilder.defineDecimal("extract", extract, minimum = Decimal.ONE)
		}

		specBuilder.pop()

		return obj
	}

	private fun batteryValues(name: String, storage: Decimal, receive: Decimal, extract: Decimal = receive, initialBatteryLevel: Decimal = Decimal.ZERO): BatteryBalanceValues {
		specBuilder.push(name)

		val obj = object : BatteryBalanceValues {
			override val capacity: Decimal by specBuilder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val receive: Decimal by specBuilder.defineDecimal("receive", receive, minimum = Decimal.ONE)
			override val extract: Decimal by specBuilder.defineDecimal("extract", extract, minimum = Decimal.ONE)
			override val initialBatteryLevel: Decimal by specBuilder.defineDecimal("initialBatteryLevel", initialBatteryLevel, minimum = Decimal.ZERO)
		}

		specBuilder.pop()

		return obj
	}

	private fun conciseValues(name: String, storage: Decimal, throughput: Decimal): ConciseBalanceValues {
		specBuilder.push(name)

		val obj = object : ConciseBalanceValues {
			override val capacity: Decimal by specBuilder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val throughput: Decimal by specBuilder.defineDecimal("throughput", throughput, minimum = Decimal.ONE)
		}

		specBuilder.pop()

		return obj
	}

	init {
		specBuilder.comment("Serverside config, holds shared values that are required to be read by both client and server.").push("server")
	}

	val LABORATORY_LAMP_LIGHT_LENGTH: Int by specBuilder.comment("In blocks").defineInRange("laboratoryLampLightLength", 6, 1, 128)

	init {
		specBuilder.comment("Energy batteries balance values").push("energyBatteries")
	}

	val BATTERY_CRUDE = batteryValues(MNames.BATTERY_CRUDE, Decimal(100_000), Decimal(160), Decimal(40), Decimal(80_000))
	val BATTERY_BASIC = batteryValues(MNames.BATTERY_BASIC, Decimal(400_000), Decimal(600))
	val BATTERY_NORMAL = batteryValues(MNames.BATTERY_NORMAL, Decimal(2_000_000), Decimal(1_000))
	val BATTERY_DENSE = batteryValues(MNames.BATTERY_DENSE, Decimal(10_000_000), Decimal(2_000))
	val BATTERY_CAPACITOR = batteryValues(MNames.BATTERY_CAPACITOR, Decimal(500_000), Decimal(50_000))

	val QUANTUM_BATTERY = conciseValues(MNames.QUANTUM_BATTERY, Decimal(40_000_000), Decimal(10_000))
	val QUANTUM_CAPACITOR = conciseValues(MNames.QUANTUM_CAPACITOR, Decimal(1_000_000), Decimal(200_000))

	val ZPM_BATTERY = conciseValues(MNames.ZPM_BATTERY, Decimal(200_000_000_000_000L), Decimal(200_000_000L))

	init {
		specBuilder.pop()

		specBuilder.comment("Matter capacitors and pattern drives balance values").push("matterCapacitorsAndDrives")
	}

	val MATTER_CAPACITOR_BASIC by specBuilder.defineDecimal(MNames.MATTER_CAPACITOR_BASIC, Decimal(2_000), minimum = Decimal.ONE_TENTH)
	val MATTER_CAPACITOR_NORMAL by specBuilder.defineDecimal(MNames.MATTER_CAPACITOR_NORMAL, Decimal(40_000), minimum = Decimal.ONE_TENTH)
	val MATTER_CAPACITOR_DENSE by specBuilder.defineDecimal(MNames.MATTER_CAPACITOR_DENSE, Decimal(400_000), minimum = Decimal.ONE_TENTH)

	val MATTER_DUST_CAPACITY by specBuilder.comment("Maximal matter value one matter dust item can have").defineDecimal("matterDustCapacity", Decimal(2_000), minimum = Decimal.ONE_TENTH)

	val PATTERN_DRIVE_NORMAL: Int by specBuilder.defineInRange(MNames.PATTERN_DRIVE_NORMAL, 4, 1, Int.MAX_VALUE)

	init {
		specBuilder.pop()

		specBuilder.comment("Balance values of machinery").push("machines")

		AndroidStationBlockEntity.registerConfig(specBuilder)
		ChemicalGeneratorBlockEntity.registerConfig(specBuilder)
		MatterRecyclerBlockEntity.registerConfig(specBuilder)
		MatterBottlerBlockEntity.registerConfig(specBuilder)
		MatterReplicatorBlockEntity.registerConfig(specBuilder)
		MatterScannerBlockEntity.registerConfig(specBuilder)
		MatterDecomposerBlockEntity.registerConfig(specBuilder)
	}

	val PLATE_PRESS = BlockEnergyStorageImpl.makeConfigEntry(specBuilder, MNames.PLATE_PRESS)
	val STORAGE_POWER_SUPPLIER = BlockEnergyStorageImpl.makeConfigEntry(specBuilder, MNames.STORAGE_POWER_SUPPLIER, capacity = Decimal(100_000), throughput = Decimal(320))
	val STORAGE_INTERFACES = BlockEnergyStorageImpl.makeConfigEntry(specBuilder, "storage_interfaces", capacity = Decimal(10_000))
	val ITEM_MONITOR = BlockEnergyStorageImpl.makeConfigEntry(specBuilder, MNames.ITEM_MONITOR)
	val DRIVE_VIEWER = BlockEnergyStorageImpl.makeConfigEntry(specBuilder, MNames.DRIVE_VIEWER)
	val DRIVE_RACK = BlockEnergyStorageImpl.makeConfigEntry(specBuilder, MNames.DRIVE_RACK, capacity = Decimal(80_000))

	init {
		specBuilder.pop()

		specBuilder.comment("Tweaking of android players").push("androidPlayer")
	}

	val REGENERATE_ENERGY: Boolean by specBuilder
		.comment("If (technically) hunger is above threshold, it turns into energy")
		.comment("This setting controls whenever to regenerate small amount of energy while eating as Android")
		.comment("And also whenever to regenerate energy while in peaceful")
		.comment("If this is disabled, any (technically) excess hunger will be nullified, unless playing on peaceful difficulty.")
		.define("regenerateEnergy", true)

	object NanobotsRegeneration {
		val COOLDOWN: List<Int> by specBuilder
			.comment("In ticks, time between heal ticks")
			.comment("One heal tick restores 1 heart (2 health points) at most")
			.comment("If not getting hurt in specified period of ticks, heal tick takes place, tick timer resets to zero and THIS array' index advances by 1")
			.comment("Index inside this array can not exceed of one of ability's")
			.comment("")
			.comment("Wording in pseudocode:")
			.comment("if (ticksSinceTakingDamage >= cooldownConfigOption[healTicks /* or config's biggest index, whichever is smaller */]) {")
			.comment("    healTicks = min(healTicks + 1, this.level /* ability level */)")
			.comment("    ticksSinceTakingDamage = 0")
			.comment("    this.ply.heal(...)")
			.comment("}")
			.defineList("cooldown", { mutableListOf(80, 60, 40, 20) }) { it is Int }

		val ENERGY_PER_HITPOINT by specBuilder
			.comment("Energy required to regenerate 1 health point (half a heart)")
			.defineDecimal("energyPerHitpoint", Decimal(800))
	}

	val ANDROID_ENERGY_PER_HUNGER_POINT by specBuilder.defineDecimal("energyPerHunger", Decimal(2000), Decimal.ZERO)
	val ANDROID_MAX_ENERGY by specBuilder.comment("Internal battery of every android has this much storage").defineDecimal("capacity", Decimal(80_000), Decimal.ZERO)
	val NIGHT_VISION_POWER_DRAW by specBuilder.defineDecimal("nightVisionPowerDraw", Decimal(8), Decimal.ZERO)
	val FALL_DAMAGE_REDUCTION_PER_LEVEL: Double by specBuilder.comment("In percent. Level of feature is multiplied by this").defineInRange("fallDamageReductionPerDampenerLevel", 0.25, 0.01, 1.0)

	object EnderTeleporter {
		init {
			specBuilder.comment("Ender Teleporter ability").push("ender_teleporter")
		}

		val ENERGY_COST by specBuilder.defineDecimal("energyCost", Decimal(4096), Decimal.ZERO)
		val COOLDOWN: Int by specBuilder.comment("In ticks").defineInRange("cooldown", 40, 0, Int.MAX_VALUE)
		val MAX_PHASE_DISTANCE: Int by specBuilder.comment("Determines how much blocks can we 'phase' through to teleport on solid surface").defineInRange("maxPhaseDistance", 6, 0, Int.MAX_VALUE)
		val MAX_DISTANCE: Double by specBuilder.comment("In blocks, euclidean distance").defineInRange("maxDistance", 12.0, 2.0, Int.MAX_VALUE.toDouble())

		init {
			specBuilder.pop()
		}
	}

	object AndroidJumpBoost {
		init {
			specBuilder.comment("Jump boost ability").push("jump_boost")
		}

		val ENERGY_COST by specBuilder.defineDecimal("energyCost", Decimal(1024), Decimal.ZERO)
		val POWER: Double by specBuilder.comment("The jump height on jump boost, as (level + 1) of feature, in meters per second").defineInRange("power", 6.0, 0.0, Double.MAX_VALUE)
		val BASE_COOLDOWN: Int by specBuilder.comment("In ticks").defineInRange("baseCooldown", 40, 0, Int.MAX_VALUE)
		val COOLDOWN_REDUCTION: Int by specBuilder.comment("In ticks, per level of feature").defineInRange("cooldownReduction", 20, 0, Int.MAX_VALUE)

		init {
			specBuilder.pop()
		}
	}

	object AndroidItemMagnet {
		init {
			specBuilder.comment("Item magnet ability").push("item_magnet")
		}

		val POWER_DRAW by specBuilder.comment("Per tick per stack").defineDecimal("powerDraw", Decimal(8), Decimal.ZERO)
		val RADIUS_HORIZONTAL: Double by specBuilder.defineInRange("radiusHorizontal", 6.0, 0.0, Double.MAX_VALUE / 4.0)
		val RADIUS_VERTICAL: Double by specBuilder.defineInRange("radiusVertical", 3.0, 0.0, Double.MAX_VALUE / 4.0)

		init {
			specBuilder.pop()
		}
	}

	object Shockwave {
		init {
			specBuilder.comment("Shockwave ability").push("shockwave")
		}

		val TERMINAL_VELOCITY: Double by specBuilder.comment("In meters per second vertically").defineInRange("terminalVelocity", 5.6, 0.0)
		val ACCELERATION: Double by specBuilder.comment("In meters per second vertically").defineInRange("acceleration", 4.0, 0.0)
		val COOLDOWN: Int by specBuilder.comment("In ticks").defineInRange("cooldown", 30, 1)
		val RADIUS_HORIZONTAL: Double by specBuilder.comment("In meters").defineInRange("radiusHorizontal", 4.0, 0.0)
		val RADIUS_VERTICAL: Double by specBuilder.comment("In meters").defineInRange("radiusVertical", 1.0, 0.0)

		val RADIUS_HORIZONTAL_WARDEN: Double by specBuilder.comment("In meters, when searching for Warden").defineInRange("radiusHorizontalWarden", 16.0, 0.0)
		val RADIUS_VERTICAL_WARDEN: Double by specBuilder.comment("In meters, when searching for Warden").defineInRange("radiusVerticalWarden", 6.0, 0.0)

		val BREAK_BLOCKS: Boolean by specBuilder.comment("Break blocks without any blast resistance").define("breakBlocks", true)
		val DAMAGE: Double by specBuilder.comment("Max potential damage done by shockwave").defineInRange("damage", 12.0, 0.0, Float.MAX_VALUE.toDouble())
		val WARDEN_DAMAGE_MULT: Double by specBuilder.defineInRange("wardenDamageMultiplier", 4.0, 0.0, Float.MAX_VALUE.toDouble())
		val ENERGY_COST by specBuilder.defineDecimal("energyCost", Decimal(2048), Decimal.ZERO)

		init {
			specBuilder.pop()
		}
	}

	init {
		// access instances so spec is built
		NanobotsRegeneration
		EnderTeleporter
		AndroidJumpBoost
		AndroidItemMagnet
		Shockwave

		specBuilder.pop()

		specBuilder.comment("Tweaking of exosuits").push("exosuitPlayer")
	}

	val INFINITE_EXOSUIT_UPGRADES: Boolean by specBuilder.comment("Allows to apply the same upgrade over and over again.", "Obviously completely breaks balance.").define("infinite_upgrades", false)

	init {
		specBuilder.pop()

		EnergySwordItem.registerConfig(specBuilder)
	}

	init {
		specBuilder.pop()
		spec = specBuilder.build()
	}

	fun register() {
		check(!registered) { "Already registered config" }
		registered = true
		ModLoadingContext.get().registerConfig(ModConfig.Type.SERVER, spec)
	}
}
