package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.IEnergyStorage
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.BatteryBankBlock
import ru.dbotthepony.mc.otm.block.IDroppableContainer
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerHooks
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.menu.BatteryBankMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities

class BatteryBankBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) : MatteryBlockEntity(MBlockEntities.BATTERY_BANK, p_155229_, p_155230_), IDroppableContainer {
	var gaugeLevel by synchronizer.float()
		private set

	// 6 на 2
	val container: MatteryContainer = object : MatteryContainer(this::setChanged, CAPACITY) {
		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			super.setChanged(slot, new, old)

			val level = level
			if (level != null) {
				var state = blockState

				for (i in BatteryBankBlock.BATTERY_SLOTS_PROPS.indices) {
					state = state.setValue(BatteryBankBlock.BATTERY_SLOTS_PROPS[i], getItem(i).getCapability(ForgeCapabilities.ENERGY).isPresent)
				}

				if (state !== blockState) {
					level.setBlock(blockPos, state, Block.UPDATE_CLIENTS)
				}
			}

			gaugeLevel = (energy.batteryLevel / energy.maxBatteryLevel).toFloat()
		}
	}

	override val droppableContainer: Container
		get() = container
	private val itemHandler = container.handler(
		object : MatteryContainerHooks {
			override fun canInsert(slot: Int, stack: ItemStack): Boolean {
				return stack.getCapability(ForgeCapabilities.ENERGY).isPresent
			}

			override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
				return true
			}
		}
	)

	private data class BatteryBankDistribution(val distribution: Array<Decimal>, val maxThroughput: Decimal)
	private enum class BankMode { RECEIVE, EXTRACT, BIDIRECTIONAL }

	private inner class BatteryBankEnergy(private val mode: BankMode) : IMatteryEnergyStorage {
		override fun canExtract() = mode != BankMode.RECEIVE
		override fun canReceive() = mode != BankMode.EXTRACT

		override fun extractEnergyOuter(howMuch: Decimal, simulate: Boolean): Decimal {
			if (mode == BankMode.RECEIVE)
				return Decimal.ZERO

			return extractEnergyInner(howMuch, simulate)
		}

		fun getDistribution(isReceiving: Boolean): BatteryBankDistribution {
			val distribution = Array(container.containerSize) { Decimal.ZERO }
			var summ = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				val stack = container.getItem(i)

				if (!stack.isEmpty) {
					stack.energy?.let {
						val diff: Decimal

						if (isReceiving) {
							diff = it.receiveEnergy(it.maxEnergyStoredMattery, true)
						} else {
							diff = it.extractEnergy(it.energyStoredMattery, true)
						}

						distribution[i] = diff
						summ += distribution[i]
					}
				}
			}

			if (!summ.isZero) {
				for (i in 0 until container.containerSize) {
					distribution[i] = distribution[i] / summ
				}
			}

			return BatteryBankDistribution(distribution, summ)
		}

		private fun distributeEnergy(isReceiving: Boolean, howMuch: Decimal, simulate: Boolean): Decimal {
			if (!howMuch.isPositive)
				return Decimal.ZERO

			val distribution = getDistribution(isReceiving)

			if (distribution.maxThroughput.isZero)
				return Decimal.ZERO

			val distList = distribution.distribution
			var summ = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				if (!distList[i].isZero) {
					val stack = container.getItem(i)

					if (!stack.isEmpty) {
						stack.energy?.let {
							val diff: Decimal

							if (isReceiving) {
								diff = it.receiveEnergy(howMuch * distList[i], simulate)
							} else {
								diff = it.extractEnergy(howMuch * distList[i], simulate)
							}

							summ += diff
						}
					}
				}
			}

			if (!simulate && !summ.isZero) {
				setChangedLight()
				gaugeLevel = (batteryLevel / maxBatteryLevel).toFloat()
			}

			return summ
		}

		override fun extractEnergyInner(howMuch: Decimal, simulate: Boolean): Decimal {
			return distributeEnergy(isReceiving = false, howMuch, simulate)
		}

		override fun receiveEnergyOuter(howMuch: Decimal, simulate: Boolean): Decimal {
			if (mode == BankMode.EXTRACT)
				return Decimal.ZERO

			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: Decimal, simulate: Boolean): Decimal {
			return distributeEnergy(isReceiving = true, howMuch, simulate)
		}

		override val batteryLevel: Decimal get() {
			var result = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				val stack = container.getItem(i)

				if (!stack.isEmpty) {
					stack.energy?.let {
						result += it.energyStoredMattery
					}
				}
			}

			return result
		}

		override val maxBatteryLevel: Decimal get() {
			var result = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				val stack = container.getItem(i)

				if (!stack.isEmpty) {
					stack.energy?.let {
						result += it.maxEnergyStoredMattery
					}
				}
			}

			return result
		}
	}

	private val energyReceiver = BatteryBankEnergy(BankMode.RECEIVE)
	private val energyExtractor = BatteryBankEnergy(BankMode.EXTRACT)
	private val energy = BatteryBankEnergy(BankMode.BIDIRECTIONAL)

	private var resolverEnergyReceiver = LazyOptional.of { energyReceiver }
	private var resolverEnergyExtractor = LazyOptional.of { energyExtractor }
	private var resolverEnergy = LazyOptional.of { energy }

	private var resolverItemHandler = LazyOptional.of { itemHandler }

	private var valid = true

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[INVENTORY_KEY] = container.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		container.deserializeNBT(nbt[INVENTORY_KEY])
		super.load(nbt)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return BatteryBankMenu(containerID, inventory, this)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverEnergyReceiver.invalidate()
		resolverEnergyExtractor.invalidate()
		resolverEnergy.invalidate()
		resolverItemHandler.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverEnergyReceiver = LazyOptional.of { energyReceiver }
		resolverEnergyExtractor = LazyOptional.of { energyExtractor }
		resolverEnergy = LazyOptional.of { energy }
		resolverItemHandler = LazyOptional.of { itemHandler }
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		tickOnceServer(this::checkSurroundings)
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap == MatteryCapability.ENERGY || cap == ForgeCapabilities.ENERGY) {
				if (side == null) return resolverEnergy.cast()

				if (side == blockState.getValue(RotatableMatteryBlock.FACING))
					return resolverEnergyExtractor.cast()
				else
					return resolverEnergyReceiver.cast()
			}

			if (cap == ForgeCapabilities.ITEM_HANDLER) {
				return resolverItemHandler.cast()
			}
		}

		return super.getCapability(cap,side)
	}

	private var consumingCapability = LazyOptional.empty<IEnergyStorage>()

	fun checkSurroundings(level: Level) {
		if (isRemoved) return

		val tile = level.getBlockEntity(blockPos.offset(blockState.getValue(RotatableMatteryBlock.FACING).normal))

		if (tile == null) {
			consumingCapability = LazyOptional.empty()
			return
		}

		consumingCapability = getAndBind(
			old = consumingCapability,
			provider = tile,
			capability = ForgeCapabilities.ENERGY,
			side = -blockState.getValue(RotatableMatteryBlock.FACING)
		) {
			@Suppress("name_shadowing")
			val level = this.level

			if (level is ServerLevel && !SERVER_IS_LIVE)
				checkSurroundings(level)
		}
	}

	fun tick() {
		if (isBlockedByRedstone)
			return

		consumingCapability.ifPresentK {
			val (_, maxThroughput) = energy.getDistribution(false)

			if (maxThroughput.isZero)
				return@ifPresentK

			val diff = it.receiveEnergy(maxThroughput, true)

			if (!diff.isZero) {
				val newExtract = energy.extractEnergyInner(diff, true)
				val newReceive = it.receiveEnergy(newExtract, true)

				val extracted = energy.extractEnergyInner(newReceive, false)
				val received = it.receiveEnergy(extracted, false)

				if (received < extracted) {
					energy.receiveEnergyInner(extracted - received, false)
				}
			}
		}
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.battery_bank")
		const val CAPACITY = 6 * 2
		private val LOGGER = LogManager.getLogger()
	}
}
