package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.block.IDroppableContainer
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.energy
import ru.dbotthepony.mc.otm.capability.energyStoredMattery
import ru.dbotthepony.mc.otm.capability.extractEnergy
import ru.dbotthepony.mc.otm.capability.maxEnergyStoredMattery
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerHooks
import ru.dbotthepony.mc.otm.core.Decimal
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.menu.EnergyServoMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MBlocks

class EnergyServoBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryBlockEntity(MBlockEntities.ENERGY_SERVO, blockPos, blockState), IDroppableContainer {
	override val defaultDisplayName: Component
		get() = MBlocks.ENERGY_SERVO.name

	val container = MatteryContainer(this::setChangedLight, 2)

	override val droppableContainer: Container
		get() = container

	val itemHandler = container.handler(object : MatteryContainerHooks {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return when (slot) {
				SLOT_DISCHARGE -> stack.isEmpty || stack.energy?.let { it.extractEnergy(Int.MAX_VALUE, true) > 0 } ?: false
				SLOT_CHARGE -> stack.isEmpty || stack.energy?.let { it.receiveEnergy(Int.MAX_VALUE, true) > 0 } ?: false
				else -> false
			}
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return when (slot) {
				SLOT_DISCHARGE -> stack.isEmpty || stack.energy?.let { it.extractEnergy(Int.MAX_VALUE, true) <= 0 } ?: false
				SLOT_CHARGE -> stack.isEmpty || stack.energy?.let { it.receiveEnergy(Int.MAX_VALUE, true) <= 0 } ?: false
				else -> false
			}
		}
	})

	val energy = object : IMatteryEnergyStorage {
		override fun extractEnergyOuter(howMuch: Decimal, simulate: Boolean): Decimal {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: Decimal, simulate: Boolean): Decimal {
			return container[SLOT_DISCHARGE].energy?.extractEnergy(howMuch, simulate) ?: Decimal.ZERO
		}

		override fun receiveEnergyOuter(howMuch: Decimal, simulate: Boolean): Decimal {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: Decimal, simulate: Boolean): Decimal {
			return container[SLOT_CHARGE].energy?.receiveEnergy(howMuch, simulate) ?: Decimal.ZERO
		}

		override val batteryLevel: Decimal
			get() = container[SLOT_CHARGE].energy?.energyStoredMattery ?: container[SLOT_DISCHARGE].energy?.energyStoredMattery ?: Decimal.ZERO

		override val maxBatteryLevel: Decimal
			get() = container[SLOT_CHARGE].energy?.maxEnergyStoredMattery ?: container[SLOT_DISCHARGE].energy?.maxEnergyStoredMattery ?: Decimal.ZERO
	}

	private var resolverEnergy = LazyOptional.of { energy }
	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverEnergy.invalidate()
		itemHandler.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverEnergy = LazyOptional.of { energy }
		itemHandler.revive()
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return EnergyServoMenu(containerID, inventory, this)
	}

	override fun <T : Any?> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			when (cap) {
				MatteryCapability.ENERGY, ForgeCapabilities.ENERGY -> return resolverEnergy.cast()
				ForgeCapabilities.ITEM_HANDLER -> return itemHandler.get().cast()
			}
		}

		return super.getCapability(cap, side)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		nbt[INVENTORY_KEY] = container.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		nbt.map(INVENTORY_KEY, container::deserializeNBT)
	}

	fun tick() {
		val charge = container[SLOT_CHARGE]
		val discharge = container[SLOT_DISCHARGE]

		if (!charge.isEmpty && !discharge.isEmpty) {
			val chargeEnergy = charge.energy ?: return
			val dischargeEnergy = discharge.energy ?: return

			val extracted = dischargeEnergy.extractEnergy(Decimal.LONG_MAX_VALUE, true)

			if (extracted.isPositive) {
				val received = chargeEnergy.receiveEnergy(extracted, true)

				if (received.isPositive) {
					val extracted2 = dischargeEnergy.extractEnergy(received, false)

					if (extracted2 == received) {
						chargeEnergy.receiveEnergy(dischargeEnergy.extractEnergy(received, false), false)
					}
				}
			}
		}
	}

	companion object {
		const val SLOT_DISCHARGE = 0
		const val SLOT_CHARGE = 1
	}
}
