package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import ru.dbotthepony.mc.otm.core.Decimal
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.data.stream

class UpdateAction : AbstractRegistryAction {
	companion object {
		private inline fun <reified T : Number> deserializeFunctionTree(input: JsonElement?, name: String): List<BoundMatterFunction<T>> {
			if (input == null) {
				return listOf()
			}

			if (input !is JsonArray) {
				throw JsonParseException("Expected $name to be JsonArray, ${input::class.qualifiedName} given")
			}

			return input.stream().map {
				if (it !is JsonObject) {
					throw JsonParseException("All elements in function tree must be JsonObjects, ${it::class.qualifiedName} given")
				}

				val type = it["type"]?.asString ?: throw JsonParseException("Invalid `type` value in function object")
				val value = it["value"] ?: throw JsonParseException("Missing `value` value in function object")

				val parsedValue: T = BoundMatterFunction.getValue(value)
				val fn: MatterFunction

				try {
					fn = MatterFunction.valueOf(type.uppercase())
				} catch (err: NoSuchElementException) {
					throw JsonParseException("No such function: $type", err)
				}

				return@map BoundMatterFunction(fn, parsedValue)
			}.collect(ImmutableList.toImmutableList())
		}
	}

	val matterFunctions: List<BoundMatterFunction<Decimal>>
	val complexityFunctions: List<BoundMatterFunction<Double>>
	val priorityFunctions: List<BoundMatterFunction<Int>>

	constructor(json: JsonObject) : super(json) {
		matterFunctions = deserializeFunctionTree(json["matter_functions"], "matter_functions")
		complexityFunctions = deserializeFunctionTree(json["complexity_functions"], "complexity_functions")
		priorityFunctions = deserializeFunctionTree(json["priority_functions"], "priority_functions")
	}

	constructor(
		tag: TagKey<Item>,
		matter: Decimal? = null,
		complexity: Double? = null,
		priority: Int? = null,
		errorOnFailure: Boolean = false,
		matterFunctions: List<BoundMatterFunction<Decimal>> = ImmutableList.of(),
		complexityFunctions: List<BoundMatterFunction<Double>> = ImmutableList.of(),
		priorityFunctions: List<BoundMatterFunction<Int>> = ImmutableList.of(),
	) : super(tag = tag, matter = matter, complexity = complexity, priority = priority, errorOnFailure = errorOnFailure) {
		check(matter != null || complexity != null || priority != null || matterFunctions.isNotEmpty() || complexityFunctions.isNotEmpty() || priorityFunctions.isNotEmpty()) {
			"Can't update existing entries without anything specified"
		}

		this.matterFunctions = ImmutableList.copyOf(matterFunctions)
		this.complexityFunctions = ImmutableList.copyOf(complexityFunctions)
		this.priorityFunctions = ImmutableList.copyOf(priorityFunctions)
	}

	constructor(
		key: ResourceLocation,
		matter: Decimal? = null,
		complexity: Double? = null,
		priority: Int? = null,
		errorOnFailure: Boolean = false,
		matterFunctions: List<BoundMatterFunction<Decimal>> = ImmutableList.of(),
		complexityFunctions: List<BoundMatterFunction<Double>> = ImmutableList.of(),
		priorityFunctions: List<BoundMatterFunction<Int>> = ImmutableList.of(),
	) : super(key = key, matter = matter, complexity = complexity, priority = priority, errorOnFailure = errorOnFailure) {
		check(matter != null || complexity != null || priority != null || matterFunctions.isNotEmpty() || complexityFunctions.isNotEmpty() || priorityFunctions.isNotEmpty()) {
			"Can't update existing entries without anything specified"
		}

		this.matterFunctions = ImmutableList.copyOf(matterFunctions)
		this.complexityFunctions = ImmutableList.copyOf(complexityFunctions)
		this.priorityFunctions = ImmutableList.copyOf(priorityFunctions)
	}

	override fun toJson(): JsonObject {
		return super.toJson().also {
			it["action"] = "update"

			if (matterFunctions.isNotEmpty()) {
				it["matter_functions"] = JsonArray().also {
					for (fn in matterFunctions) {
						it.add(fn.toJson())
					}
				}
			}

			if (complexityFunctions.isNotEmpty()) {
				it["complexity_functions"] = JsonArray().also {
					for (fn in complexityFunctions) {
						it.add(fn.toJson())
					}
				}
			}

			if (priorityFunctions.isNotEmpty()) {
				it["priority_functions"] = JsonArray().also {
					for (fn in priorityFunctions) {
						it.add(fn.toJson())
					}
				}
			}
		}
	}

	private fun apply(value: MutableEntry, modifier: ResourceLocation) {
		if (matterFunctions.isNotEmpty())
			value.matter = matterFunctions.apply(value.matter)
		else if (matter != null)
			value.matter = matter

		if (complexityFunctions.isNotEmpty())
			value.complexity = complexityFunctions.apply(value.complexity)
		else if (complexity != null)
			value.complexity = complexity

		if (priorityFunctions.isNotEmpty())
			value.priority = priorityFunctions.apply(value.priority)
		else if (priority != null)
			value.priority = priority
	}

	override fun update(registry: MutableCollection<MutableEntry>, modifier: ResourceLocation) {
		if(!(
			matter != null ||
			complexity != null ||
			priority != null ||
			matterFunctions.isNotEmpty() ||
			complexityFunctions.isNotEmpty() ||
			priorityFunctions.isNotEmpty()
		)) {
			throw JsonParseException("Can't update existing entries without anything specified (for $modifier)")
		}

		if (matter != null && matterFunctions.isNotEmpty()) {
			throw JsonParseException("Can't have both matter and matter functions specified at the same time (for $modifier)")
		}

		if (complexity != null && complexityFunctions.isNotEmpty()) {
			throw JsonParseException("Can't have both complexity and complexity functions specified at the same time (for $modifier)")
		}

		if (priority != null && priorityFunctions.isNotEmpty()) {
			throw JsonParseException("Can't have both priority and priority functions specified at the same time (for $modifier)")
		}

		if (!checkConditions()) {
			return
		}

		if (key != null) {
			val iterator = registry.iterator()

			for (value in iterator) {
				if (value is MutableKeyEntry && value.key == key) {
					apply(value, modifier)
					return
				}
			}

			fail { "Could not find matter value with key $key" }
		} else {
			val iterator = registry.iterator()

			for (value in iterator) {
				if (value is MutableTagEntry && value.tag == tag) {
					apply(value, modifier)
					return
				}
			}

			fail { "Could not find matter value with tag $tag" }
		}
	}
}
